import pytest
import os
import sys
import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import gext
import gext.grassmann
import utils

SMALL = 1e-10

@pytest.mark.parametrize("datafile", ["urea.json", "glucose.json"])
def test_grassmann(datafile):

    # load test data from json file
    data = utils.load_json(f"tests/{datafile}")
    nelectrons = data["nelectrons"]
    natoms = data["trajectory"].shape[1]
    nbasis = data["overlaps"].shape[1]
    nframes = data["trajectory"].shape[0]

    # initialize an extrapolator
    extrapolator = gext.Extrapolator(nelectrons, nbasis, natoms, nframes)

    # load data in the extrapolator
    for (coords, coeff, overlap) in zip(data["trajectory"],
            data["coefficients"], data["overlaps"]):
        extrapolator.load_data(coords, coeff, overlap)

    # check the Grassmann projections
    c0 = extrapolator._get_tangent()
    for (coeff, gamma, overlap) in zip(data["coefficients"],
            extrapolator.gammas.get(nframes), extrapolator.overlaps.get(nframes)):

        sqrt_overlap = extrapolator._sqrt_overlap(overlap)

        coeff = sqrt_overlap @ coeff[:, :nelectrons]
        d = coeff @ coeff.T

        # standard checks on the density
        assert np.linalg.norm(d - d.T, ord=np.inf) < SMALL
        assert np.linalg.norm(d - d @ d, ord=np.inf) < SMALL
        assert np.trace(d) - nelectrons < SMALL

        # compute the density from the inverse Grassmann map: Exp(Log(D))
        coeff_1 = gext.grassmann.exp(gamma, c0)
        d_1 = coeff_1 @ coeff_1.T

        # standard check on the second density
        assert np.linalg.norm(d_1 - d_1.T, ord=np.inf) < SMALL
        assert np.linalg.norm(d_1 - d_1 @ d_1, ord=np.inf) < SMALL
        assert np.trace(d_1) - nelectrons < SMALL

        # finally check if Exp(Log(D)) is an identity
        assert np.linalg.norm(d - d_1, ord=np.inf)
